/* Copyright (c) 1997-2009
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Darmstadt, Germany)
   http://www.polymake.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
--------------------------------------------------------------------------------
   $Project: polymake $$Id: cayley_polytope.cc 9527 2010-03-04 17:52:36Z herr $
*/

#include <polymake.h>
#include <Integer.h>
#include <Rational.h>
#include <Matrix.h>
#include <ListMatrix.h>
#include <Vector.h>

namespace polymake { namespace polytope {

   perl::Object cayley_polytope(const Array<perl::Object> & pp, perl::OptionSet options)
   {
      int d = pp.size();

      Entire< Array<perl::Object> >::const_iterator pp_it = entire(pp);

      if ( !pp_it->give("LATTICE") )
	 throw std::runtime_error("cayley polytope construction only defined for lattice polytopes");

      perl::Object p_out(pp_it->type());
      std::string names = pp_it->name();
      ListMatrix< Vector<Integer> > U = pp_it->give("VERTICES | POINTS");
      U |= repeat_row(unit_vector<Integer>(d,0),U.rows());

      int i = 1;
      while ( !(++pp_it).at_end() ) {
	 if ( !pp_it->give("LATTICE") )
	    throw std::runtime_error("cayley polytope construction only defined for lattice polytopes");

	 Matrix<Integer> V = pp_it->give("VERTICES | POINTS");
	 U /= V|repeat_row(unit_vector<Integer>(d,i++),V.rows());

	 names += ", ";
	 names += pp_it->name();
      }

      p_out.set_description() << "cayley polytope of the polytopes " << names << endl;
      if ( options["proj"] ) 
	 p_out.take("POINTS") << U.minor(All,~range(U.cols()-1,U.cols()-1));
      else
	 p_out.take("POINTS") << U;

      return p_out;
   }

   UserFunctionTemplate4perl("# @category Producing a new polyhedron from others"
			     "# Construct the cayley polytope of a set of lattice polytopes contained in //P_Array//"
			     "# which is the convex hull of P<sub>1</sub>&times;e<sub>1</sub>, ..., P<sub>k</sub>&times;e<sub>k</sub>"
			     "# where e<sub>1</sub>, ...,e<sub>k</sub> are the standard unit vectors in R<sup>k</sup>."
			     "# In this representation the last k coordinates always add up to 1."
			     "# The option //proj// projects onto the complement of the last coordinate."
			     "# @param Array<LatticePolytope> P_Array  an array containing the lattice polytopes P<sub>1</sub>,...,P<sub>k</sub>"
			     "# @option Bool proj"
			     "# @return Polytope",
			     "cayley_polytope(Polytope<Rational> +; {proj => 0} )");
} }
