/* Copyright (c) 1997-2009
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Darmstadt, Germany)
   http://www.polymake.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
--------------------------------------------------------------------------------
   $Project: polymake $$Id: wreath.cc 9523 2010-03-04 13:27:19Z herr $
*/

#include <vector>
#include <Rational.h>
#include <polymake.h>
#include <Matrix.h>
#include <Set.h>
#include <IncidenceMatrix.h>
#include <sstream>

namespace polymake { namespace polytope {

namespace {

struct wreath_label {
   typedef std::string first_argument_type;
   typedef std::string second_argument_type;
   typedef std::string result_type;
    
   result_type operator() (first_argument_type l1, second_argument_type l2) const
   {
      return l2 + '^' + l1;
   }
};

template <typename Coord>
Matrix<Coord>
wreath_coord(const Matrix<Coord>& V1, const Matrix<Coord>& V2)
{
   const int n_vertices1(V1.rows()), n_vertices2(V2.rows()), dim1(V1.cols()-1), dim2(V2.cols()-1), d(n_vertices2*dim1+dim2+1);
   Matrix<Coord> V(n_vertices1*n_vertices2,d);
    
   for (int i=0; i<n_vertices2; ++i) {
      for (int k=0; k<n_vertices1; ++k) {
	const int ik(i*n_vertices1+k);
	V(ik,0)=Rational(1);
	const int id1(i*dim1); 
	int j(1);

	//v_{ik,j} = 0 for 0 <= j <= k*dim1
	for ( ; j<=id1; ++j) V(ik,j)=0;

	//v_{ik,j} = v1_{i, j-k*dim1} for k*dim1 < j <= (k+1)*dim1
	for ( ; j<=id1+dim1; ++j) V(ik,j)=V1(k,j-id1)/V1(k,0);

	// v_{ik,j} = 0 for (k+1)*dim1 < j <= n_vertices*dim1
	for ( ; j<=n_vertices2*dim1; ++j) V(ik,j)=0;
	
	//v_{ik,j} = v2_{k,j-n_vertices2*dim1} for n_vertices2*dim1 < j < d
	for ( ; j<d; ++j) V(ik,j)=V2(i,j-n_vertices2*dim1)/V2(i,0);
      }
   }
   return V;
}
  
} // end unnamed namespace

template <typename Coord>
  perl::Object wreath(perl::Object p_in1, perl::Object p_in2, perl::OptionSet options)
{
  const bool dual=options["dual"];
  const bool relabel=options["relabel"];

  const bool centered1 = p_in1.give("CENTERED");
  if (!centered1) {
    throw std::runtime_error("first polytope must be CENTERED.");
  }
  const bool centered2 = p_in2.give("CENTERED");
  if (!centered2) {
    throw std::runtime_error("second polytope must be CENTERED.");
  }
  
  std::ostringstream section;
  std::ostringstream labels_section;
  if (!dual) {
    section << "VERTICES";
    labels_section << "VERTEX";
  } else {
    section << "FACETS";
    labels_section << "FACET";
  }
  const Matrix<Coord> M1=p_in1.give(section.str().c_str()),
    M2=p_in2.give(section.str().c_str());
  const int n_rows1(M1.rows()), n_rows2(M2.rows()),
    n_rows_out(n_rows1*n_rows2);
  
  const Matrix<Coord> M_out = wreath_coord<Coord>(M1, M2);
  

  perl::Object p_out("Polytope");
  p_out.take(section.str().c_str()) << M_out;
  if (relabel) {
    std::vector<std::string> labels1(n_rows1), labels2(n_rows2),
      labels_out(n_rows_out);
    labels_section << "_LABELS";
    read_labels(p_in1, labels_section.str().c_str(), labels1);
    read_labels(p_in2, labels_section.str().c_str(), labels2);
    std::copy(entire(product(labels2, labels1, wreath_label())), labels_out.begin());
    p_out.take(labels_section.str().c_str()) << labels_out;
  }

  return p_out;
}
  UserFunctionTemplate4perl("# @category Producing a new polyhedron from others"
			    "# Construct a new polytope as the wreath product of two input polytopes //P1//, //P2//."
			    "# //P1// and //P2// have to be [[BOUNDED]]."
			    "# @param Polytope P1"
			    "# @param Polytope P2"
			    "# @option Bool dual invokes the computation of the dual wreath product"
			    "# @option Bool relabel creates an additional section [[VERTEX_LABELS]];" 
			    "#   the label of a new vertex corresponding to v<sub>1</sub> &oplus; v<sub>2</sub> will" 
			    "#   have the form LABEL_1*LABEL_2."
			    "# @return Polytope",
			    "wreath<Coord>(Polytope<Coord> Polytope<Coord> { dual => 0, relabel => 0})");
} }
