/* Copyright (c) 1997-2009
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Darmstadt, Germany)
   http://www.polymake.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
--------------------------------------------------------------------------------
   $Project: polymake $$Id: minimal_non_faces.cc 9298 2009-09-02 23:10:56Z gawrilow $
*/

#include <polymake.h>
#include <PowerSet.h>
#include <Integer.h>
#include <HasseDiagram.h>
#include <Bitset.h>
#include <list>
#include <hash_set>

namespace polymake { namespace topaz {
  
void minimal_non_faces(perl::Object p)
{
   const HasseDiagram HD = p.give("HASSE_DIAGRAM");

   const int dim = HD.dim()-1;  // dimension of the complex = HD.dim()-1
   std::list< Set<int> > min_non_faces;

   // determine the start level (the highest complete level)
   int start_dim = 1;
   while ( HD.nodes_of_dim(start_dim).size() == Integer::binom(dim,start_dim) )  ++start_dim;
   --start_dim;

   const int n_vertices=HD.node_range_of_dim(0).size(), top_node=HD.top_node();

   // iterate over all levels of HD and determine the minimal non-faces one above
   for (int d=start_dim; d<=dim; ++d) {

      // create hash set containing all faces of this dimension
      hash_set< Set<int> > faces(HD.node_range_of_dim(d).size());
      for (Entire<sequence>::const_iterator it=entire(HD.node_range_of_dim(d)); !it.at_end(); ++it)
	 faces.insert(HD.face(*it));
      const hash_set< Set<int> >::iterator faces_end = faces.end();  

      // iterate over all faces of this dimension
      for (Entire<sequence>::const_iterator it=entire(HD.node_range_of_dim(d)); !it.at_end(); ++it) {
	 const int n=*it;
	 const Set<int>& f= HD.face(n);
	 
	 Bitset non_candidates(n_vertices);
	 for (Entire<Graph<Directed>::out_edge_list>::const_iterator e=entire(HD.out_edges(n)); !e.at_end(); ++e)
	    if (e.to_node() != top_node)
	       non_candidates += (HD.face(e.to_node())-f).front();
#ifdef POLY_DEBUG
	 cout << "\nface " << n << ": " << f << "\nnon_candidates: " << non_candidates << endl;
#endif
	 // generate non-faces
	 for (Entire<Graph<Directed>::in_edge_list>::const_iterator in_e=entire(HD.in_edges(n)); !in_e.at_end(); ++in_e) {
	    for (Entire<Graph<Directed>::out_edge_list>::const_iterator out_e=entire(HD.out_edges(in_e.from_node())); !out_e.at_end(); ++out_e) {
	       const int nn = out_e.to_node();
	       if (n==nn)
		  continue;

	       const int candidate = (HD.face(nn)-f).front();
	       if ( non_candidates.contains(candidate) || candidate<f.back() )
		  continue;
#ifdef POLY_DEBUG
	       cout << "candidate node " << nn << " (" << HD.face(nn) << ") with vertex: " << candidate << endl;
#endif
	       // test the candidate
	       bool is_minimal = true;
	       for (Entire< Subsets_less_1< Set<int> > >::const_iterator f_it=entire(all_subsets_less_1(f)); !f_it.at_end(); ++f_it)
		  if ( faces.find(*f_it + candidate) == faces_end ) {
		     is_minimal = false;
		     break;
		  }
	       if (is_minimal)  min_non_faces.push_back(f + candidate);
	       non_candidates += candidate;
#ifdef POLY_DEBUG
	       if (is_minimal)  cout << "ADDING min non-face: " << f + candidate << endl;
#endif
	    }
	 }

      }  // end iterate over all faces of this dimension
   }

   p.take("MINIMAL_NON_FACES") << min_non_faces;
}

   Function4perl(&minimal_non_faces, "minimal_non_faces(SimplicialComplex)");
} }
